package lcsb.mapviewer.web.api.project.map.bioentity.reactions;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import javax.validation.Valid;
import javax.validation.constraints.NotNull;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.hibernate.validator.constraints.NotBlank;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import lcsb.mapviewer.api.BaseController;
import lcsb.mapviewer.api.QueryException;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.reaction.Modifier;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.modelutils.map.ElementUtils;
import lcsb.mapviewer.persist.dao.map.ReactionProperty;
import lcsb.mapviewer.persist.dao.map.species.ElementProperty;
import lcsb.mapviewer.services.ObjectNotFoundException;
import lcsb.mapviewer.services.interfaces.IElementService;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IProjectService;
import lcsb.mapviewer.services.interfaces.IReactionService;
import lcsb.mapviewer.web.api.NewApiResponseSerializer;

@RestController
@RequestMapping(
    value = {
        "/minerva/new_api/projects/{projectId}/models/{mapId}/bioEntities/reactions/" },
    produces = MediaType.APPLICATION_JSON_VALUE)
public class NewReactionsController extends BaseController {

  private IReactionService reactionService;
  private IElementService elementService;
  private IProjectService projectService;
  private IModelService modelService;
  private NewApiResponseSerializer serializer;

  @Autowired
  public NewReactionsController(final IElementService elementService, final IProjectService projectService, final IModelService modelService,
      final IReactionService reactionService,
      final NewApiResponseSerializer serializer) {
    this.reactionService = reactionService;
    this.projectService = projectService;
    this.modelService = modelService;
    this.serializer = serializer;
    this.elementService = elementService;
  }

  @PreAuthorize("hasAnyAuthority('IS_ADMIN','READ_PROJECT:' + #projectId)")
  @GetMapping(value = "{reactionId}")
  public ResponseEntity<?> listReactions(
      final @Valid @NotBlank @PathVariable(value = "projectId") String projectId,
      final @Valid @NotNull @PathVariable(value = "mapId") String mapId,
      final @Valid @NotBlank @PathVariable(value = "reactionId") String reactionId) throws QueryException, ObjectNotFoundException {

    if (!StringUtils.isNumeric(reactionId)) {
      throw new QueryException("Invalid reactionId: " + reactionId);
    }
    if (!StringUtils.isNumeric(mapId)) {
      throw new QueryException("Invalid mapId: " + mapId);
    }
    Map<ReactionProperty, List<? extends Object>> properties = new HashMap<>();
    properties.put(ReactionProperty.ID, Arrays.asList(new Integer(reactionId)));

    properties.put(ReactionProperty.PROJECT, Arrays.asList(projectService.getProjectByProjectId(projectId)));
    if (!mapId.equals("*")) {
      try {
        properties.put(ReactionProperty.MAP, Arrays.asList(modelService.getModelByMapId(projectId, Integer.valueOf(mapId))));
      } catch (NumberFormatException e) {
        throw new QueryException("mapId must be numeric", e);
      }
    }

    Page<Reaction> reactions = reactionService.getReactionsByFilter(properties, true);
    if (reactions.getContent().size() == 0) {
      throw new ObjectNotFoundException("Reaction does not exists");
    }
    return serializer.prepareResponse(reactions.getContent().get(0));
  }

  @PreAuthorize("hasAnyAuthority('IS_ADMIN','READ_PROJECT:' + #projectId)")
  @PostMapping(value = ":downloadCsv")
  public ResponseEntity<?> downloadCsv(
      final @Valid @NotBlank @PathVariable(value = "projectId") String projectId,
      final @Valid @NotNull @PathVariable(value = "mapId") String mapId,
      final @Valid @NotBlank @RequestBody DownloadCsvRequest request)
      throws QueryException, IOException, NumberFormatException, ObjectNotFoundException {

    Map<ElementProperty, List<? extends Object>> properties = new HashMap<>();
    properties.put(ElementProperty.PROJECT, Arrays.asList(projectService.getProjectByProjectId(projectId)));
    if (request.getElementTypes().size() > 0) {
      properties.put(ElementProperty.CLASS, new ElementUtils().getClassesByStringTypes(request.getElementTypes()));
    }
    if (NumberUtils.isCreatable(mapId)) {
      properties.put(ElementProperty.MAP, Arrays.asList(modelService.getModelByMapId(projectId, Integer.valueOf(mapId))));
    } else if (request.getSubmaps().size() > 0) {
      List<ModelData> models = modelService.getModelsByProjectId(projectId);
      List<ModelData> filteredModels = new ArrayList<>();
      for (ModelData modelData : models) {
        if (request.getSubmaps().contains(modelData.getId())) {
          filteredModels.add(modelData);
        }
      }

      if (filteredModels.size() == 0) {
        throw new QueryException("Submaps don't exist");
      }
      properties.put(ElementProperty.MAP, filteredModels);
    }
    if (request.getIncludedCompartmentIds().size() > 0) {
      properties.put(ElementProperty.INCLUDED_IN_COMPARTMENT, request.getIncludedCompartmentIds());
    }
    if (request.getExcludedCompartmentIds().size() > 0) {
      properties.put(ElementProperty.EXCLUDED_FROM_COMPARTMENT, request.getExcludedCompartmentIds());
    }

    List<Element> filteredElements = elementService.getElementsByFilter(properties);

    Set<Integer> elementIds = new HashSet<>();
    for (Element element : filteredElements) {
      elementIds.add(element.getId());
    }
    List<Reaction> reactions;
    if (elementIds.size() > 0) {
      reactions = reactionService.getReactionById(projectId, "*", new HashSet<>(), elementIds);
    } else {
      reactions = new ArrayList<>();
    }

    List<Reaction> filteredReactions = new ArrayList<>();
    if (request.getReactionTypes().size() > 0) {
      for (Reaction reaction : reactions) {
        if (getStriungReactionTypes(request.getReactionTypes()).contains(reaction.getStringType())) {
          filteredReactions.add(reaction);
        }
      }
    } else {
      filteredReactions = reactions;
    }

    String content = prepareCsv(filteredReactions, request.getColumns(), request.getAnnotations());
    byte[] out = content.getBytes();

    HttpHeaders responseHeaders = new HttpHeaders();
    responseHeaders.add("content-disposition", "attachment; filename=elements.csv");
    responseHeaders.add("Content-Type", "text/csv");

    return new ResponseEntity<>(out, responseHeaders, HttpStatus.OK);
  }

  private Set<String> getStriungReactionTypes(final List<String> reactionTypes) throws QueryException {
    Set<String> result = new HashSet<>();
    for (String string : reactionTypes) {
      if (string.equals("Process description")) {
        result.addAll(Arrays.asList("Dissociation",
            "Heterodimer association",
            "Inhibition",
            "Known transition omitted",
            "Modulation",
            "Physical stimulation",
            "State transition",
            "Transcription",
            "Translation",
            "Transport",
            "Trigger",
            "Truncation",
            "Unknown catalysis",
            "Unknown inhibition",
            "Unknown transition"));
      } else if (string.equals("Activity flow")) {
        result.addAll(Arrays.asList("Reduced modulation",
            "Reduced physical stimulation",
            "Reduced trigger",
            "Unknown reduced modulation",
            "Unknown reduced physical stimulation",
            "Unknown reduced trigger",
            "Negative influence",
            "Positive influence",
            "Unknown negative influence",
            "Unknown positive influence"));
      } else {
        throw new QueryException("Unvalid reaction type: " + string);
      }
    }
    return result;
  }

  private String prepareCsv(final List<Reaction> reactions, final List<String> columns, final List<String> annotations)
      throws IOException, QueryException {
    List<String> header = new ArrayList<>(columns);
    header.addAll(annotations);
    Writer writer = new StringWriter();
    try (CSVPrinter csvPrinter = new CSVPrinter(writer, CSVFormat.DEFAULT
        .builder().setHeader(header.toArray(new String[header.size()])).build())) {
      for (Reaction element : reactions) {
        csvPrinter.printRecord(getCsvRow(element, columns, annotations));
      }

      csvPrinter.flush();
      return writer.toString();
    }
  }

  private List<Object> getCsvRow(final Reaction reaction, final List<String> columns, final List<String> annotations) throws QueryException {
    List<Object> result = new ArrayList<>();
    Object value;
    for (String string : columns) {
      value = null;
      switch (string) {
        case ("id"):
          value = reaction.getId();
          break;
        case ("description"):
          value = reaction.getNotes();
          break;
        case ("modelId"):
          value = reaction.getModelData().getId();
          break;
        case ("mapName"):
          value = reaction.getModelData().getName();
          break;
        case ("symbol"):
          value = reaction.getSymbol();
          break;
        case ("abbreviation"):
          value = reaction.getAbbreviation();
          break;
        case ("synonyms"):
          value = StringUtils.join(reaction.getSynonyms(), ";");
          break;
        case ("references"):
          List<String> refs = new ArrayList<>();
          for (MiriamData md : reaction.getMiriamData()) {
            refs.add(md.getDataType() + ":" + md.getResource());
          }
          value = StringUtils.join(refs, ";");
          break;
        case ("name"):
          value = reaction.getName();
          break;
        case ("type"):
          value = reaction.getStringType();
          break;
        case ("elements"):
          List<String> participants = new ArrayList<>();
          for (Reactant reactant : reaction.getReactants()) {
            participants.add("REACTANT:" + reactant.getElement().getId());
          }
          for (Product product : reaction.getProducts()) {
            participants.add("PRODUCT:" + product.getElement().getId());
          }
          for (Modifier modifier : reaction.getModifiers()) {
            participants.add("MODIFIER:" + modifier.getElement().getId());
          }
          value = StringUtils.join(participants, ";");
          break;
        case ("reactantIds"):
          List<String> reactants = new ArrayList<>();
          for (Reactant reactant : reaction.getReactants()) {
            reactants.add(reactant.getElement().getId() + "");
          }
          value = StringUtils.join(reactants, ";");
          break;
        case ("productIds"):
          List<String> products = new ArrayList<>();
          for (Product product : reaction.getProducts()) {
            products.add(product.getElement().getId() + "");
          }
          value = StringUtils.join(products, ";");
          break;
        case ("modifierIds"):
          List<String> modifiers = new ArrayList<>();
          for (Modifier modifier : reaction.getModifiers()) {
            modifiers.add(modifier.getElement().getId() + "");
          }
          value = StringUtils.join(modifiers, ";");
          break;
        case ("reactionId"):
          value = reaction.getIdReaction();
          break;
        case ("mechanicalConfidenceScore"):
          value = reaction.getMechanicalConfidenceScore();
          break;
        case ("lowerBound"):
          value = reaction.getLowerBound();
          break;
        case ("upperBound"):
          value = reaction.getUpperBound();
          break;
        case ("geneProteinReaction"):
          value = reaction.getGeneProteinReaction();
          break;
        case ("subsystem"):
          value = reaction.getSubsystem();
          break;
        default:
          throw new QueryException("Unknown column name: " + string);
      }
      result.add(value);
    }
    for (String annotation : annotations) {
      value = null;
      try {
        MiriamType type = MiriamType.valueOf(annotation);
        List<String> refs = new ArrayList<>();
        for (MiriamData md : reaction.getMiriamData()) {
          if (Objects.equals(md.getDataType(), type)) {
            refs.add(md.getResource());
          }
        }
        value = StringUtils.join(refs, ";");
        result.add(value);
      } catch (IllegalArgumentException e) {
        throw new QueryException("Invalid annotation name: " + annotation, e);
      }
    }

    return result;
  }

}